/**
 * 
 */
package uk.org.wuglug.gryle.util;

import java.io.IOException;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

/**
 * @author rlmw
 *
 */
public class JavaSoundMusicPlayer extends AbstractMusicPlayer {

	private JSMPThread thread = null;

	private boolean isPaused = false;

	private boolean stop = false;
	
	private AudioInputStream din;
	private SourceDataLine line;
	
	private Object lock = new Object();

	//-----------------------------------
	//  Publically accessible methods 
	//-----------------------------------
	
	/**
	 * @throws IOException 
	 * @see uk.org.wuglug.gryle.util.MusicPlayer#pause()
	 */
	public void pause() throws IOException {
		if (this.thread != null) {
			synchronized (lock) {
				if(!this.isPaused) {
					this.isPaused = true;
					stop();
				} else {
					this.lock.notify();
				}
			}
		}		
	}

	/**
	 * @see uk.org.wuglug.gryle.util.MusicPlayer#play()
	 */
	public void play() {
		this.thread = new JSMPThread(this);
		this.thread.start();
	}
	
	/**
	 * @throws IOException 
	 * @see uk.org.wuglug.gryle.util.MusicPlayer#stop()
	 */
	public void stop() throws IOException {
		synchronized (lock) {
			if (this.thread != null) {
				line.drain();
				line.stop();
				line.close();
				din.close();
			}
		}		
	}
	
	//-----------------------------------
	//  Methods used by the internal thread
	//-----------------------------------

	public void startPlaying() {
		try {
			//AudioFileFormat aff = AudioSystem.getAudioFileFormat(this.in);
			AudioInputStream in = AudioSystem.getAudioInputStream(this.in);
			if (in != null) {
				AudioFormat baseFormat = in.getFormat();
				AudioFormat decodedFormat = new AudioFormat(
						AudioFormat.Encoding.PCM_SIGNED, baseFormat
								.getSampleRate(), 16, baseFormat.getChannels(),
						baseFormat.getChannels() * 2, baseFormat
								.getSampleRate(), false);
				din = AudioSystem.getAudioInputStream(decodedFormat, in);
				rawplay(decodedFormat);
				in.close();
			}
		} catch (UnsupportedAudioFileException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (LineUnavailableException e) {
			e.printStackTrace();
		}
	}

	private SourceDataLine getLine(AudioFormat audioFormat)
			throws LineUnavailableException {
		SourceDataLine res = null;
		DataLine.Info info = new DataLine.Info(SourceDataLine.class,
				audioFormat);
		res = (SourceDataLine) AudioSystem.getLine(info);
		res.open(audioFormat);
		return res;
	}

	private void rawplay(AudioFormat targetFormat)
			throws IOException, LineUnavailableException {
		
		byte[] data = new byte[4096];
		line = getLine(targetFormat);
		if (line != null) {
			// Start
			line.start();
			int nPreviousBytesRead = 0, nBytesRead = 0, nBytesWritten = 0;
			while (nBytesRead != -1) {
				nBytesRead = pauseAndReinitIfNecessary(targetFormat,nPreviousBytesRead,nBytesRead);
				nPreviousBytesRead = nBytesRead;
				nBytesRead = din.read(data, 0, data.length);
				if (nBytesRead != -1)
					nBytesWritten = line.write(data, 0, nBytesRead);
				
			}
			// Stop
			stop();
		}
	}

	private int pauseAndReinitIfNecessary(AudioFormat targetFormat, int nPreviousBytesRead, int nBytesRead) throws LineUnavailableException {
		synchronized (lock) {
			if (this.isPaused) {
				try {
					this.lock.wait();
				} catch (InterruptedException e) {}
				// now reinitiatise the data
				line = getLine(targetFormat);
			}
		}
		return nBytesRead;
	}

}

class JSMPThread extends Thread {

	@Override
	public void run() {
		this.jsmp.startPlaying();
	}

	private JavaSoundMusicPlayer jsmp;

	public JSMPThread(JavaSoundMusicPlayer jsmp) {
		this.jsmp = jsmp;
	}
}
